"""
Core functions for building nests.
"""

import collections
import errno
import functools
import json
import os
import os.path
import warnings

from ._py3 import is_string, imap

CONTROL_NAME = 'control.json'

# Load a JSON file into an ordered dict
ordered_load = functools.partial(json.load,
        object_pairs_hook=collections.OrderedDict)

def stripext(path):
    """
    Return the basename, minus extension, of a path.

    :param string path: Path to file
    """
    return os.path.basename(os.path.splitext(path)[0])

def _mkdirs(d):
    """
    Make all directories up to d.

    No exception is raised if d exists.
    """
    try:
        os.makedirs(d)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

_Nestable = collections.namedtuple('Nestable', ('name', 'nestable',
                                               'create_dir', 'update',
                                               'label_func'))

def _is_iter(iterable):
    """
    Return whether an item is iterable or not
    """
    try:
        iter(iterable)
        return True
    except TypeError:
        return False

def _repeat_iter(iterable):
    def repeat_iter(ctl):
        return iterable
    return repeat_iter

def _templated(fn):
    """
    Return a function which applies ``str.format(**ctl)`` to all results of
    ``fn(ctl)``.
    """
    @functools.wraps(fn)
    def inner(ctl):
        return [i.format(**ctl) for i in fn(ctl)]
    return inner

class Nest(object):
    """
    Nests are used to build nested parameter selections, culminating in a
    directory structure representing choices made, and a JSON dictionary with
    all selections.

    Build parameter combinations with :meth:`Nest.add`, then create a nested
    directory structure with :meth:`Nest.build`.

    :param control_name: Name JSON file to be created in each leaf
    :param indent: Indentation level in json file
    :param fail_on_clash: Error if a nest level attempts to overwrite a
        previous value
    :param warn_on_clash: Print a warning if a nest level attempts ot overwrite
        a previous value
    :param base_dict: Base dictionary to start all control dictionaries from
        (default: ``{}``)
    :param include_outdir: If true, include an OUTDIR key in every control
        indicating the directory this control would be written to.
    """
    def __init__(self, control_name=CONTROL_NAME, indent=2,
            fail_on_clash=False, warn_on_clash=True, base_dict=None,
            include_outdir=True):
        self.control_name = control_name
        self.indent = indent
        self.fail_on_clash = fail_on_clash
        self.warn_on_clash = warn_on_clash
        self.include_outdir = include_outdir
        if base_dict is None:
            base_dict = collections.OrderedDict()
        if self.include_outdir:
            base_dict['OUTDIR'] = ''
        self._controls = [('', base_dict)]

    def iter(self, root=None):
        """
        Create an iterator of (directory, control_dict) tuples for all valid
        parameter choices in this :class:`Nest`.

        :param root: Root directory
        :rtype: Generator of ``(directory, control_dictionary)`` tuples.
        """
        if root is None:
            return iter(self._controls)
        return ((os.path.join(root, outdir), control)
                for outdir, control in self._controls)

    def __iter__(self):
        """
        Iterate over directory, control pairs. Same as :meth:`Nest.iter`.
        """
        return self.iter()

    def build(self, root="runs"):
        """
        Build a nested directory structure, starting in ``root``

        :param root: Root directory for structure
        """

        for d, control in self.iter(root):
            _mkdirs(d)
            with open(os.path.join(d, self.control_name), 'w') as fp:
                json.dump(control, fp, indent=self.indent)
                # RJSON and some other tools like a trailing newline
                fp.write('\n')

    def add(self, name, nestable, create_dir=True, update=False,
            label_func=str, template_subs=False):
        """
        Add a level to the nest

        :param string name: Name of the level. Forms the key in the output
            dictionary.
        :param nestable: Either an iterable object containing values, _or_ a
            function which takes a single argument (the control dictionary)
            and returns an iterable object containing values
        :param boolean create_dir: Should a directory level be created for this
            nestable?
        :param boolean update: Should the control dictionary be updated with
            the results of each value returned by the nestable? Only valid for
            dictionary results; useful for updating multiple values. At a
            minimum, a key-value pair corresponding to ``name`` must be
            returned.
        :param label_func: Function to be called to convert each value to a
            directory label.
        :param boolean template_subs: Should the strings in / returned by
            nestable be treated as templates? If true, str.format is called
            with the current values of the control dictionary.
        """
        # Convert everything to functions
        if not callable(nestable):
            if not _is_iter(nestable):
                raise ValueError("Invalid nestable: " + str(nestable))
            if is_string(nestable):
                warnings.warn(
                        "Passed a string as an iterable for name {0}".format(name))
            old_nestable = nestable
            nestable = _repeat_iter(old_nestable)
        if template_subs:
            nestable = _templated(nestable)

        new_controls = []
        for outdir, control in self._controls:
            for r in nestable(control):
                new_outdir, new_control = outdir, control.copy()
                if update:
                    # Make sure expected key exists
                    if name not in r:
                        raise KeyError("Missing key for {0}".format(name))
                    # Check for collisions
                    u = frozenset(control.keys()) & frozenset(r.keys())
                    if u:
                        msg = "Key overlap: {0}".format(u)
                        if self.fail_on_clash:
                            raise KeyError(msg)
                        elif self.warn_on_clash:
                            warnings.warn(msg)
                    new_control.update(r)
                    to_label = r[name]
                else:
                    new_control[name] = to_label = r

                if create_dir:
                    new_outdir = os.path.join(outdir, label_func(to_label))
                if self.include_outdir:
                    new_control['OUTDIR'] = new_outdir
                new_controls.append((new_outdir, new_control))

        self._controls = new_controls


def nest_map(control_iter, map_fn):
    """
    Apply ``map_fn`` to the directories defined by ``control_iter``

    For each control file in control_iter, map_fn is called with the directory
    and control file contents as arguments.

    Example::

        >>> list(nest_map(['run1/control.json', 'run2/control.json'],
        ...               lambda d, c: c['run_id']))
        [1, 2]

    :param control_iter: Iterable of paths to JSON control files
    :param function map_fn: Function to run for each control file. It should
            accept two arguments: the directory of the control file and the
            json-decoded contents of the control file.
    :returns: A generator of the results of applying ``map_fn`` to elements in
            ``control_iter``
    """
    def fn(control_path):
        """
        Read the control file, return the result of calling map_fn
        """
        with open(control_path) as fp:
            control = ordered_load(fp)
        dn = os.path.dirname(control_path)
        return map_fn(dn, control)

    mapped = imap(fn, control_iter)
    return mapped

def control_iter(base_dir, control_name=CONTROL_NAME):
    """
    Generate the names of all control files under base_dir
    """
    controls = (os.path.join(p, control_name) for p, _, fs in os.walk(base_dir)
                if control_name in fs)
    return controls



"""SCons integration for nestly."""
from collections import OrderedDict
import json
import logging
import copy
import os
import warnings

from . import core

try:
    import SCons.Node
    import SCons.Node.FS
    HAS_SCONS = True
except ImportError:
    warnings.warn('Unable to import SCons. Some functionality not available.')
    HAS_SCONS = False

logger = logging.getLogger('nestly.scons')

class SConsEncoder(json.JSONEncoder):
    """
    JSON Encoder which handles SCons objects.
    """
    def default(self, obj):
        if isinstance(obj, SCons.Node.NodeList):
            return list(obj)
        elif isinstance(obj, (SCons.Node.FS.Entry, SCons.Node.FS.File)):
            return str(obj)
        return super(SConsEncoder, self).default(obj)

def _create_control_file(source, target, env):
    target = str(target[0])
    with open(target, 'w') as fp:
        json.dump(env['control_dict'], fp, indent=2, cls=env['encoder_cls'])

def name_targets(func):
    """
    Wrap a function such that returning ``'a', 'b', 'c', [1, 2, 3]`` transforms
    the value into ``dict(a=1, b=2, c=3)``.

    This is useful in the case where the last parameter is an SCons command.
    """
    def wrap(*a, **kw):
        ret = func(*a, **kw)
        return dict(zip(ret[:-1], ret[-1]))
    return wrap


class SConsWrap(object):
    """A Nest wrapper to add SCons integration.

    This class wraps a :class:`Nest <nestly.core.Nest>` in order to provide
    methods which are useful for using nestly with SCons.

    A Nest passed to SConsWrap must have been created with
    ``include_outdir=True``, which is the default.

    :param nest: A :class:`Nest <nestly.core.Nest>` object to wrap
    :param dest_dir: The base directory for all output directories.
    :param alias_environment: An optional SCons ``Environment`` object.
        If present, targets added via :meth:`SConsWrap.add_target` will include
        an alias using the nest key.
    """

    def __init__(self, nest, dest_dir='.', alias_environment=None):
        """Initialize an SConsWrap.

        Takes the Nest to operate on and the base directory for all output
        directories.
        """
        self.nest = nest
        self.dest_dir = dest_dir
        self.alias_environment = alias_environment
        self.checkpoints = OrderedDict()

    def __iter__(self):
        "Iterate over the current controls."
        return self.nest.iter(self.dest_dir)

    def add(self, name, nestable, **kw):
        """
        Adds a level to the nesting and creates a checkpoint that can be
        reverted to later for aggregation by calling :meth:`SConsWrap.pop`.

        :param name: Identifier for the nest level
        :param nestable: A nestable object - see
            :meth:`Nest.add() <nestly.core.Nest.add>`.
        :param kw: Additional parameters to pass to
            :meth:`Nest.add() <nestly.core.Nest.add>`.
        """
        self.checkpoints[name] = self.nest
        self.nest = copy.copy(self.nest)
        return self.nest.add(name, nestable, **kw)

    def pop(self, name=None):
        """
        Reverts to the nest stage just before the corresponding call of
        :meth:`SConsWrap.add_aggregate`.  However, any aggregate collections
        which have been worked on will still be accessible, and can be called
        operated on together after calling this method.  If no name is passed,
        will revert to the last nest level.

        :param name: Name of the nest level to pop.
        """
        if name is not None:
            self.nest = self.checkpoints[name]
            keys = list(self.checkpoints.keys())
            name_idx = keys.index(name)
            assert name_idx >= 0

            # Pop every key from ``name`` on:
            for k in reversed(keys[name_idx:]):
                self.checkpoints.pop(k)
        else:
            self.nest = self.checkpoints.popitem()[1]

    def add_nest(self, name=None, **kw):
        """A simple decorator which wraps :meth:`nestly.core.Nest.add`."""
        def deco(func):
            self.add(name or func.__name__, func, **kw)
            return func
        return deco

    def _register_alias(self, key):
        if self.alias_environment:
            values = [c[key] for _, c in self if c[key]]
            values = self.alias_environment.Flatten(values)
            if values and any(isinstance(v, dict) for v in values):
                warnings.warn(('Skipping adding alias for {0}: '
                               'dictionaries in output.').format(key))
                return
            if values and not any(isinstance(v, dict) for v in values):
                self.alias_environment.Alias(key, values)

    def add_target(self, name=None):
        """
        Add an SCons target to this nest.

        The function decorated will be immediately called with each of the
        output directories and current control dictionaries. Each result will
        be added to the respective control dictionary for later nests to
        access.

        :param name: Name for the target in the name (default: function name).
        """
        def deco(func):
            def nestfunc(control):
                destdir = os.path.join(self.dest_dir, control['OUTDIR'])
                return [func(destdir, control)]
            key = name or func.__name__
            self.nest.add(key, nestfunc, create_dir=False)
            self._register_alias(key)
            return func
        return deco

    def add_target_with_env(self, environment, name=None):
        """Add an SCons target to this nest, with an SCons Environment

        The function decorated will be immediately called with three arguments:

        * ``environment``: A clone of the SCons environment, with variables
          populated for all values in the control dictionary, plus a variable
          ``OUTDIR``.
        * ``outdir``: The output directory
        * ``control``: The control dictionary

        Each result will be added to the respective control dictionary for
        later nests to access.

        Differs from :meth:`SConsWrap.add_target` only by the addition of the
        ``Environment`` clone.
        """
        def deco(func):
            def nestfunc(control):
                env = environment.Clone()
                for k, v in control.items():
                    if k in env:
                        logger.warn("Overwriting previously bound value %s=%s",
                                    k, env[k])
                    env[k] = v
                destdir = os.path.join(self.dest_dir, control['OUTDIR'])
                env['OUTDIR'] = destdir
                return [func(env, destdir, control)]
            key = name or func.__name__
            self.nest.add(key, nestfunc, create_dir=False)
            self._register_alias(key)
            return func
        return deco

    def add_aggregate(self, name, data_fac):
        """
        Add an aggregate target to this nest.


        Since nests added after the aggregate can access the construct returned
        by the factory function value, it can be mutated to provide additional
        values for use when the decorated function is called.

        To do something with the aggregates, you must :meth:`SConsWrap.pop`
        nest levels created between addition of the aggregate and then can add
        any normal targets you would like which take advantage of the targets
        added to the data structure.

        :param name: Name for the target in the nest
        :param data_fac: a nullary factory function which will be called
            immediately for each of the current control dictionaries and stored
            in each dictionary with the given name as in
            :meth:`SConsWrap.add_target`.
        """
        @self.add_target(name)
        def wrap(outdir, c):
            return data_fac()

    def add_controls(self, env, target_name='control',
                     file_name='control.json',
                     encoder_cls=SConsEncoder):
        """
        Adds a target to build a control file at each of the current leaves.

        :param env: SCons Environment object
        :param target_name: Name for target in nest
        :param file_name: Name for output file.
        """
        if not HAS_SCONS:
            raise ImportError('SCons not available')

        @self.add_target(name=target_name)
        def control(outdir, c):
            return env.Command(os.path.join(outdir, file_name),
                               [],
                               action=_create_control_file,
                               control_dict=c,
                               encoder_cls=encoder_cls)



"""
Utilities for dealing with python3
"""

import itertools
import sys

py3 = sys.version_info[0] == 3

if py3:
    imap = map
else:
    imap = itertools.imap

def is_string(s):
    if py3:
        b = str
    else:
        b = basestring
    return isinstance(s, b)



import copy
import unittest
import mock

from nestly import scons, Nest

class OutputCopyingMock(mock.MagicMock):
    def __call__(self, *args, **kwargs):
        return copy.deepcopy(super(OutputCopyingMock, self).__call__(*args, **kwargs))

class AddTargetWithEnvTestCase(unittest.TestCase):
    def setUp(self):
        # Default output will a return a reference, which means consecutive
        # calls to .Clone will return references to the same dictionary. Use
        # OutputCopyingMock to avoid clobbering mutation.
        self.env = OutputCopyingMock(['Clone'], name='MockSConsEnvironment')
        self.env.Clone.return_value = {}
        self.func_mock = mock.Mock('__call__', name='func_mock')
        self.func_mock.__name__ = 'func_mock'
        self.n = Nest()
        self.n.add('item', [1, 2])

    def test_basic(self):
        w = scons.SConsWrap(self.n)

        w.add_target_with_env(self.env)(self.func_mock)

        # Clone is be called with no arguments
        self.env.Clone.assert_called_with()

        calls = [mock.call({'item': 1, 'OUTDIR': './1'}, './1', {'item': 1, 'OUTDIR': '1'}),
                 mock.call({'item': 2, 'OUTDIR': './2'}, './2', {'item': 2, 'OUTDIR': '2'})]
        self.func_mock.assert_has_calls(calls)

class CheckpointTestCase(unittest.TestCase):
    def setUp(self):
        self.nest = Nest()
        self.nest.add('list', [[]], create_dir=False)
        self.w = scons.SConsWrap(self.nest)
        self.w.add('level1', range(2))
        self.w.add('level2', (1, 2, 3))

    def test_pop_with_name(self):
        w = self.w

        @w.add_target()
        def key_file(outdir, c):
            c['list'].append(c['level2'])
            return True

        @w.add_target()
        def assertion_during(outdir, c):
            self.assertTrue(c['key_file'])
            self.assertTrue('level2' in c)
            return True

        n2 = w.nest
        w.pop('level1')
        n1 = w.nest

        # Want to make sure nested level checkpoints are popped off as well
        self.assertEqual(len(w.checkpoints), 0)
        # Make sure we have the right nest now
        self.assertTrue(self.nest is n1)
        self.assertFalse(self.nest is n2)

        @w.add_target()
        def assertion_after(outdir, c):
            self.assertFalse('key_file' in c)
            self.assertEqual(6, len(c['list']))
            self.assertFalse('level2' in c)

    def test_pop_last(self):
        w = self.w
        w.pop()

        # This time, there should be one checkpoint left
        self.assertEqual(len(w.checkpoints), 1)
        self.assertTrue('level1' in w.checkpoints)

        @w.add_target()
        def assertion_after(outdir, c):
            self.assertFalse('level2' in c)
            self.assertTrue('level1' in c)

    def test_pop_missing(self):
        self.assertRaises(KeyError, self.w.pop, 'missing_key')
        self.assertEqual(['level1', 'level2'],
                         sorted(self.w.checkpoints.keys()))


def suite():
    suite = unittest.TestSuite()
    for cls in [AddTargetWithEnvTestCase, CheckpointTestCase]:
        suite.addTest(unittest.makeSuite(cls))
    return suite



"""
nestrun.py - run commands based on control dictionaries.
"""
import argparse
import collections
import csv
import datetime
import errno
import functools
import json
import logging
import os
import os.path
import shlex
import shutil
import signal
import subprocess
import sys

from nestly.core import control_iter

# Constants to be used as defaults.
MAX_PROCS = 2                    # Set the default maximum number of child processes that can be spawned.
DRY_RUN = False                   # Run in dry_run mode, default is False.


def _terminate_procs(procs):
    """
    Terminate all processes in the process dictionary
    """
    logging.warn("Stopping all remaining processes")
    for proc, g in procs.values():
        logging.debug("[%s] SIGTERM", proc.pid)
        try:
            proc.terminate()
        except OSError as e:
            # we don't care if the process we tried to kill didn't exist.
            if e.errno != errno.ESRCH:
                raise
    sys.exit(1)

def sigterm_handler(nlocal, signum, frame):
    logging.warning('SIGTERM received; no longer spawning jobs')
    nlocal['spawn_jobs'] = False

def sigusr1_handler(running_procs, signum, frame):
    for pid, (proc, _) in running_procs.iteritems():
        sys.stderr.write('%5d - in %s\n' % (pid, proc.working_dir))
    sys.stderr.flush()  # just in case it's being buffered by something

def sigint_handler(nlocal, write_this_summary, running_procs, signum, frame):
    if nlocal['received_SIGINT']:
        logging.warning('SIGINT received; terminating')
        _terminate_procs(running_procs)
        write_this_summary()
        sys.exit(0)
    else:
        logging.warning('SIGINT received; send again to terminate')
        nlocal['received_SIGINT'] = True

def invoke(max_procs, data, json_files):
    nlocal = {'spawn_jobs': True, 'received_SIGINT': False}
    running_procs = {}
    all_procs = []
    def write_this_summary():
        write_summary(all_procs, data['summary_file'])

    signal.signal(signal.SIGTERM, functools.partial(sigterm_handler, nlocal))
    signal.signal(signal.SIGUSR1,
        functools.partial(sigusr1_handler, running_procs))
    signal.signal(signal.SIGINT,
        functools.partial(sigint_handler, nlocal, write_this_summary,
                          running_procs))

    files = iter(json_files)
    try:
        while True:
            while nlocal['spawn_jobs'] and len(running_procs) < max_procs:
                try:
                    json_file = files.next()
                except StopIteration:
                    # no more files; allow other processes to finish.
                    break
                g = worker(data, json_file)
                try:
                    proc = g.next()
                except StopIteration:
                    continue
                except OSError:
                    # OSError thrown when command couldn't be started
                    logging.exception("Exception starting %s", json_file)
                    if data['stop_on_error']:
                        _terminate_procs(running_procs)
                        return
                else:
                    all_procs.append(proc)
                    running_procs[proc.pid] = proc, g

            try:
                pid, status = os.wait()
            except OSError as e:
                # wait(2) raising ECHILD means there's no child processes to wait
                # for anymore, so we're done.
                if e.errno == errno.ECHILD:
                    return
                elif e.errno == errno.EINTR:
                    continue
                else:
                    raise

            # Pull the actual exit status - high byte of 16-bit number
            exit_status = os.WEXITSTATUS(status)
            proc, g = running_procs.pop(pid)
            proc.complete(exit_status)

            try:
                g.next()
            except StopIteration:
                pass
            else:
                raise ValueError('worker generators should only yield once')

            # Check exit status, cancel jobs if stop_on_error specified and
            # non-zero
            if exit_status:
                logging.warn('[%s] %s Finished with non-zero exit status %s\n%s',
                        pid, proc.working_dir, exit_status, proc.log_tail())
                if data['stop_on_error']:
                    _terminate_procs(running_procs)
                    return
            else:
                logging.info("[%s] %s Finished with %s", pid, proc.working_dir,
                        exit_status)
    finally:
        write_this_summary()


def write_summary(all_procs, summary_file):
    """
    Write a summary of all run processes to summary_file in tab-delimited
    format.
    """
    if not summary_file:
        return

    with summary_file:
        writer = csv.writer(summary_file, delimiter='\t', lineterminator='\n')
        writer.writerow(('directory', 'command', 'start_time', 'end_time',
            'run_time', 'exit_status', 'result'))
        rows = ((p.working_dir, ' '.join(p.command), p.start_time, p.end_time,
                p.running_time, p.return_code, p.status) for p in all_procs)
        writer.writerows(rows)


def template_subs_file(in_file, out_fobj, d):
    """
    Substitute template arguments in in_file from variables in d, write the
    result to out_fobj.
    """
    with open(in_file, 'r') as in_fobj:
        for line in in_fobj:
            out_fobj.write(line.format(**d))


class NestlyProcess(object):
    """
    Metadata about a process run
    """

    def __init__(self, command, working_dir, popen, log_name='log.txt'):
        self.command = command
        self.working_dir = working_dir
        self.log_name = log_name
        self.popen = popen
        self.pid = popen.pid
        self.return_code = None
        self.start_time = datetime.datetime.now()
        self.end_time = None
        self.status = 'RUNNING'

    def terminate(self):
        self.popen.terminate()
        self.end_time = datetime.datetime.now()
        self.status = 'TERMINATED'

    def complete(self, return_code):
        """
        Mark the process as complete with provided return_code
        """
        self.return_code = return_code
        self.status = 'COMPLETE' if not return_code else 'FAILED'
        self.end_time = datetime.datetime.now()

    @property
    def running_time(self):
        if self.end_time is None:
            return None

        return self.end_time - self.start_time

    def log_tail(self, nlines=10):
        """
        Return the last ``nlines`` lines of the log file
        """
        log_path = os.path.join(self.working_dir, self.log_name)
        with open(log_path) as fp:
            d = collections.deque(maxlen=nlines)
            d.extend(fp)
            return ''.join(d)


def worker(data, json_file):
    """
    Handle parameter substitution and execute command as child process.
    """
    # PERHAPS TODO: Support either full or relative paths.
    with open(json_file) as fp:
        d = json.load(fp)
    json_directory = os.path.dirname(json_file)
    def p(*parts):
        return os.path.join(json_directory, *parts)

    # STDOUT and STDERR will be written to a log file in each job directory.
    log_file = data['log_file']

    # A template file will be written in each job directory, including the
    # substitution that was performed..
    savecmd_file = data['savecmd_file']

    # if a template file is being used, then we write out to it
    template_file = data['template_file']
    if template_file:
        output_template = p(os.path.basename(template_file))
        with open(output_template, 'w') as out_fobj:
            template_subs_file(template_file, out_fobj, d)

        # Copy permissions to destination
        try:
            shutil.copymode(template_file, output_template)
        except OSError as e:
            if e.errno == errno.EPERM:
                logging.warn("Couldn't set permissions on %s. "
                        "Continuing with existing permissions",
                        output_template)
            else:
                raise

    work = data['template'].format(**d)

    if savecmd_file:
        with open(p(savecmd_file), 'w') as command_file:
            command_file.write(work + "\n")

    # View what actions will take place in dry_run mode.
    if data['dry_run']:
        logging.info("%s - Dry run of %s\n", p(), work)
    else:
        try:
            with open(p(log_file), 'w') as log:
                cmd = shlex.split(work)
                while True:
                    try:
                        pr = subprocess.Popen(
                            cmd, stdout=log, stderr=log, cwd=p())
                    except OSError as e:
                        if e.errno != errno.EINTR:
                            raise
                        continue
                    else:
                        break
                logging.info('[%d] Started %s in %s', pr.pid, work, p())
                nestproc = NestlyProcess(cmd, p(), pr)
                yield nestproc
        except Exception as e:
            # Seems useful to print the command that failed to make the
            # traceback more meaningful.  Note that error output could get
            # mixed up if two processes encounter errors at the same instant
            logging.error("%s - Error executing %s - %s", p(), work, e)
            raise e


def extant_file(x):
    """
    'Type' for argparse - checks that file exists but does not open.
    """
    if not os.path.exists(x):
        raise argparse.ArgumentError("{0} does not exist".format(x))
    return x


def parse_arguments():
    """
    Grab options and json files.
    """
    max_procs = MAX_PROCS
    dry_run = DRY_RUN
    logging.basicConfig(level=logging.INFO, stream=sys.stdout,
                        format='%(asctime)s * %(levelname)s * %(message)s')

    parser = argparse.ArgumentParser(description="""nestrun - substitute values
            into a template and run commands in parallel.""")
    parser.add_argument('-j', '--processes', '--local', dest='local_procs',
            type=int, help="""Run a maximum of N processes in parallel locally
            (default: %(default)s)""", metavar='N', default=MAX_PROCS)
    parser.add_argument('--template', dest='template',
            metavar="'template text'", help="""Command-execution template, e.g.
            bash {infile}. By default, nestrun executes the templatefile.""")
    parser.add_argument('--stop-on-error', action='store_true',
            default=False, help="""Terminate remaining processes if any process
            returns non-zero exit status (default: %(default)s)""")
    parser.add_argument('--template-file', dest='template_file', metavar="FILE",
            help='Command-execution template file path.')
    parser.add_argument('--save-cmd-file', dest='savecmd_file',
            help="""Name of the file that will contain the command that was
            executed.""")
    log_group = parser.add_mutually_exclusive_group()
    log_group.add_argument('--log-file', dest='log_file', default='log.txt',
            help="""Name of the file that will contain output of the executed
            command.""")
    log_group.add_argument('--no-log', dest="log_file", action="store_const",
            default='log.txt', const=os.devnull, help="""Don't create a log
            file""")
    parser.add_argument('--dry-run', action='store_true', help="""Dry run mode,
            does not execute commands.""", default=False)
    parser.add_argument('--summary-file', type=argparse.FileType('w'),
            help="""Write a summary of the run to the specified file""")

    ctrl_group = parser.add_argument_group('Control files')
    ctrl_group.add_argument('json_files', metavar='control_files', type=extant_file,
            nargs='*', help="""Nestly control dictionaries""")
    ctrl_group.add_argument('-d', '--directory', help="""Run on all control
            files under %(metavar)s. May be used in place of specifying control
            files.""", metavar='DIR')
    arguments = parser.parse_args()


    # Load controls
    if bool(arguments.directory) == bool(arguments.json_files):
        parser.error('Exactly one of `-d` and control_files must be specified.')
    elif arguments.directory:
        arguments.json_files.extend(control_iter(arguments.directory))

    template = arguments.template

    # Make sure that either a template or a template file was given
    if arguments.template_file:
        # if given a template file, the default is to run the input
        if not arguments.template:
            template = os.path.join('.',
                    os.path.basename(arguments.template_file))

            # If using the default argument, the template must be executable:
            if (not os.access(arguments.template_file, os.X_OK) and not
                    arguments.dry_run):
                parser.error(
                        "{0} is not executable. Specify a template.".format(
                    arguments.template_file))

    if not (arguments.template or arguments.template_file):
        parser.exit("Error: Please specify either a template "
                "or a template file")

    logging.info('Template: %s', template)

    if arguments.local_procs is not None:
        max_procs = arguments.local_procs

    # Create a dictionary that will be shared amongst all forked processes.
    data = {}
    data['dry_run'] = arguments.dry_run
    data['start_directory'] = os.getcwd()
    data['template'] = template
    data['template_file'] = arguments.template_file
    data['savecmd_file'] = arguments.savecmd_file
    data['log_file'] = arguments.log_file
    data['stop_on_error'] = arguments.stop_on_error
    data['summary_file'] = arguments.summary_file

    return data, max_procs, arguments.json_files

def main():
    data, max_procs, json_files = parse_arguments()
    invoke(max_procs, data, json_files)




"""
Aggregate results of ``nestly`` runs.
"""

import argparse
import collections
import csv
import functools
import itertools
import os.path
import json
import sys

from ..core import control_iter, nest_map

DEFAULT_SEP = ','
DEFAULT_NAME = 'control.json'

# JSON loaders retaining key order
_ordered_load = functools.partial(json.load,
                                  object_pairs_hook=collections.OrderedDict)
_ordered_loads = functools.partial(json.loads,
                                   object_pairs_hook=collections.OrderedDict)

def warn(message):
    print >>sys.stderr, message

def _warn_on_io(fn):
    @functools.wraps(fn)
    def f(*args, **kwargs):
        r = fn(*args, **kwargs)
        try:
            for i in r:
                yield i
        except IOError as e:
            warn(str(e))
    return f

def _delim_accum(control_files, filename_template, keys=None,
        exclude_keys=None, separator=DEFAULT_SEP, missing_action='fail'):
    """
    Accumulator for delimited files

    Combines each file with values from JSON dictionary in same directory

    :param iterable control_files: Iterable of control files
    :param filename_template: A template for the file to nest_map
    :param keys: List of keys to select from JSON dictionary. If ``None``, keep
                 all keys.
    :param separator: Delimiter
    """
    def map_fn(d, control, keys=keys):
        f = os.path.join(d, filename_template.format(**control))

        keys = keys if keys is not None else control.keys()
        if exclude_keys:
            keys = list(frozenset(keys) - frozenset(exclude_keys))
        if frozenset(keys) - frozenset(control):
            # Unknown keys
            raise ValueError(
                    "The following required key(s) are not present in {1}: {0}".format(
                        ', '.join(frozenset(keys) - frozenset(control)),
                        f))
        with open(f) as fp:
            reader = csv.DictReader(fp, delimiter=separator)
            for row in reader:
                row_dict = collections.OrderedDict(
                        itertools.chain(((k, row[k]) for k in reader.fieldnames),
                        ((k, v) for k, v in control.items() if k in keys)))

                yield row_dict
    if missing_action == 'warn':
        map_fn = _warn_on_io(map_fn)

    return itertools.chain.from_iterable(nest_map(control_files, map_fn))

def delim(arguments):
    """
    Execute delim action.

    :param arguments: Parsed command line arguments from :func:`main`
    """

    if bool(arguments.control_files) == bool(arguments.directory):
        raise ValueError(
                'Exactly one of control_files and `-d` must be specified.')

    if arguments.directory:
        arguments.control_files.extend(control_iter(arguments.directory))

    with arguments.output as fp:
        results = _delim_accum(arguments.control_files,
                arguments.file_template, arguments.keys,
                arguments.exclude_keys, arguments.separator,
                missing_action=arguments.missing_action)
        r = next(results)
        writer = csv.DictWriter(fp, r.keys(), delimiter=arguments.separator)
        writer.writeheader()
        writer.writerow(r)
        writer.writerows(results)

def comma_separated_values(s):
    s = s.split(',')
    return s

def main(args=sys.argv[1:]):
    """
    Command-line interface for nestagg
    """
    parser = argparse.ArgumentParser(description="""Aggregate results of
            nestly runs""")
    subparsers = parser.add_subparsers()
    delim_parser = subparsers.add_parser('delim', help="""Combine control files
            with delimited files.""")
    delim_parser.set_defaults(func=delim)
    key_group = delim_parser.add_mutually_exclusive_group()
    key_group.add_argument('-k', '--keys', help="""Comma separated list of
            keys from the JSON file to include [default: all keys]""",
            type=comma_separated_values)
    key_group.add_argument('-x', '--exclude-keys', help="""Comma separated
            list of keys from the JSON file not to include [default:
            %(default)s]""", type=comma_separated_values)
    delim_parser.add_argument('-m', '--missing-action', choices=('fail',
        'warn'), help="""Action to take when a file is missing [default:
        %(default)s]""", default='fail')
    delim_parser.add_argument('file_template', help="""Template for the
            delimited file to read in each directory [e.g. '{run_id}.csv']""")
    delim_parser.add_argument('control_files', metavar="control.json",
            help="""Control files""", nargs="*")
    delim_parser.add_argument('-d', '--directory', help="""Run on all control
            files under %(metavar)s. May be used in place of specifying control
            files.""", metavar='DIR')
    delim_parser.add_argument('-s', '--separator', default=DEFAULT_SEP,
            help="""Separator [default: %(default)s]""")
    delim_parser.add_argument('-t', '--tab', action='store_const',
            dest='separator', const='\t', help="""Files are tab-separated""")
    delim_parser.add_argument('-o', '--output', default=sys.stdout,
        type=argparse.FileType('w'), help="""Output file [default: stdout]""")

    arguments = parser.parse_args()

    arguments.func(arguments)



import unittest

from . import test_core, test_scons

def suite():
    suite = unittest.TestSuite()
    for mod in [test_core, test_scons]:
        suite.addTest(mod.suite())
    return suite

suite = suite()



import contextlib
import json
import os
import os.path
import unittest
import tempfile
import shutil
import warnings

from nestly import core

@contextlib.contextmanager
def tempdir():
    td = tempfile.mkdtemp()
    try:
        yield td
    finally:
        shutil.rmtree(td)

class NestCompareMixIn(object):
    def assertNestsEqual(self, expected, actual):
        self.assertEqual(len(expected), len(actual))
        for (ie, de), (ia, da) in zip(expected, actual):
            self.assertEqual(ie, ia)
            self.assertEqual(de, da)

class SimpleNestTestCase(NestCompareMixIn, unittest.TestCase):

    def setUp(self):
        nest = core.Nest(include_outdir=False)
        nest.add("number", (1, 10))
        nest.add("name", ("a", "b"))
        self.nest = nest
        self.expected =  [('1/a', {'name': 'a', 'number': 1}),
                          ('1/b', {'name': 'b', 'number': 1}),
                          ('10/a', {'name': 'a', 'number': 10}),
                          ('10/b', {'name': 'b', 'number': 10})]

    def test_iter_once(self):
        actual = list(self.nest.iter())
        self.assertNestsEqual(self.expected, actual)

    def test_iter_repeatable(self):
        # Run once
        list(self.nest.iter())
        actual = list(self.nest.iter())
        self.assertNestsEqual(self.expected, actual)

    def test_iter_prefix(self):
        actual = list(self.nest.iter('test2/test'))
        expected = [('test2/test/' + a, b) for a, b in self.expected]
        self.assertNestsEqual(expected, actual)

    def test_build(self):
        with tempdir() as td:
            self.nest.build(td)
            actual = [os.path.join(p, f) for p, d, files in os.walk(td)
                      for f in files]
            expected = {os.path.join(td, a, 'control.json'): b
                        for a, b in self.expected}

            # Test that all controls were created
            self.assertEqual(frozenset(expected.keys()), frozenset(actual))

            for a in actual:
                with open(a) as fp:
                    d = json.load(fp)
                self.assertEqual(expected[a], d)

    def test_stringiter_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.nest.add('string', 'value')
            self.assertEqual(1, len(w))


class TemplateTestCase(NestCompareMixIn, unittest.TestCase):
    """
    Test template substitution
    """
    def test_template(self):
        nest = core.Nest(include_outdir=False)
        nest.add('number', [1, 2])
        nest.add('dirname', ['number-{number}'], template_subs=True,
                create_dir=False)
        actual = list(nest.iter())
        expected = [('1', {'dirname': 'number-1', 'number': 1}),
                    ('2', {'dirname': 'number-2', 'number': 2})]
        self.assertNestsEqual(expected, actual)


class UpdateTestCase(NestCompareMixIn, unittest.TestCase):

    def test_update(self):
        nest = core.Nest(include_outdir=False)
        values = [{'number': 1, 'description': 'one'},
                  {'number': 2, 'description': 'two'}]
        nest.add("number", values, update=True)
        actual = list(nest.iter())
        expected = list(zip(('1', '2'), values))
        self.assertNestsEqual(expected, actual)

    def test_update_nokey(self):
        nest = core.Nest(include_outdir=False)
        self.assertRaises(KeyError, nest.add,
                          "number", [{'description': 'one'}], update=True)

    def test_update_overwrite(self):
        nest = core.Nest(fail_on_clash=True, include_outdir=False)
        nest.add("description", ['Test'])
        values = [{'number': 1, 'description': 'one'},
                  {'number': 2, 'description': 'two'}]
        self.assertRaises(KeyError, nest.add, "number", values, update=True)


class OutdirTestCase(NestCompareMixIn, unittest.TestCase):

    def test_outdir(self):
        nest = core.Nest(include_outdir=True)
        nest.add('a', [1, 2])
        nest.add('b', [3, 4])
        actual = list(nest.iter())
        expected = [
            ('1/3', dict(a=1, b=3, OUTDIR='1/3')),
            ('1/4', dict(a=1, b=4, OUTDIR='1/4')),
            ('2/3', dict(a=2, b=3, OUTDIR='2/3')),
            ('2/4', dict(a=2, b=4, OUTDIR='2/4')),
        ]
        self.assertNestsEqual(expected, actual)


class IsIterTestCase(unittest.TestCase):

    def test_list(self):
        self.assertTrue(core._is_iter([1, 2, 3]))

    def test_generator(self):
        g = (i for i in range(4))
        self.assertTrue(core._is_iter(g))

        # Can't consume
        self.assertEqual([0, 1, 2, 3], list(g))

    def test_non_iterable(self):
        non_iters = [False, True, 9, 4.5, object()]
        for i in non_iters:
            self.assertFalse(core._is_iter(i))

class SimpleNestMixin(object):
    """
    Builds a temporary nest
    """
    def setUp(self):
        self.td = tempfile.mkdtemp(prefix='nest')
        n = core.Nest()
        n.add('run_id', (1, 2))
        n.build(self.td)

        self.controls = [os.path.join(p, f)
                         for p, _, files in os.walk(self.td)
                         for f in files if f == 'control.json']

    def tearDown(self):
        shutil.rmtree(self.td)

    def test_preconditions(self):
        self.assertEqual(2, len(self.controls))

class NestMapTestCase(SimpleNestMixin, unittest.TestCase):
    def test_provides_dirs(self):
        actual = sorted(core.nest_map(self.controls, lambda d, c: d))
        expected = sorted(os.path.dirname(f) for f in self.controls)
        self.assertEqual(expected, actual)

    def test_control(self):
        actual = sorted(core.nest_map(self.controls, lambda d, c: c['run_id']))
        expected = [1, 2]
        self.assertEqual(expected, actual)

def suite():
    suite = unittest.TestSuite()
    for cls in [IsIterTestCase,
            NestMapTestCase,
            SimpleNestTestCase,
            TemplateTestCase,
            UpdateTestCase]:
        suite.addTest(unittest.makeSuite(cls))
    return suite
